//  ICE Revision: $Id: ValueExpressionDriver.C 7103 2007-02-01 21:53:52Z bgschaid $ 

#include "ValueExpressionDriver.H"
#include "ValueExpressionParser.tab.hh"
#include <Random.H>
#include <wallDist.H>
#include "zeroGradientFvPatchFields.H"
#include "fixedValueFvPatchFields.H"
#include "fvPatchFieldFields.H"
#include "wallFvPatch.H"
#include "cellSet.H"

ValueExpressionDriver::ValueExpressionDriver (const Foam::string &time,const Foam::fvMesh &mesh)
    : time(time),
      mesh(mesh),
      result(NULL),
      vresult(NULL),
      typ(NO_TYPE),
      trace_scanning (false), 
      trace_parsing (false)
{
}

ValueExpressionDriver::~ValueExpressionDriver ()
{
    if(result) {
        delete result;
    }
    if(vresult) {
        delete vresult;
    }
}

void ValueExpressionDriver::setScalarResult(Foam::volScalarField *r) {
    result=r;
    typ=SCALAR_TYPE;
}

void ValueExpressionDriver::setLogicalResult(Foam::volScalarField *r) {
    result=r;
    typ=LOGICAL_TYPE;
}

void ValueExpressionDriver::setVectorResult(Foam::volVectorField *r) {
    vresult=r;
    typ=VECTOR_TYPE;
}

void ValueExpressionDriver::parse (const std::string &f)
{
    content = f;
    scan_begin ();
    yy::ValueExpressionParser parser (*this);
    parser.set_debug_level (trace_parsing);
    parser.parse ();
    scan_end ();
}

void ValueExpressionDriver::error (const yy::location& l, const std::string& m)
{
    std::ostringstream buff;
    buff << l;
    std::string place="";
    for(unsigned int i=0;i<l.begin.column;i++) {
        place+=" ";
    }
    for(unsigned int i=l.begin.column;i<l.end.column;i++) {
        place+="^";
    }
    for(unsigned int i=l.end.column;i<content.size();i++) {
        place+=" ";
    }

    Foam::FatalErrorIn("parsingValue")
        //        << Foam::args.executable()
        << " Parser Error at " << buff.str() << " :"  << m << Foam::endl
            << content << Foam::endl << place
            << Foam::exit(Foam::FatalError);
    //    Foam::Info << buff.str() << ": " << m << Foam::endl;
}

void ValueExpressionDriver::error (const std::string& m)
{
    Foam::FatalErrorIn("parsingValue")
        //        << Foam::args.executable()
            << " Parser Error: " << m
            << Foam::exit(Foam::FatalError);
}

bool ValueExpressionDriver::isFaceScalar(const Foam::string &name)
{
    if(getTypeOfField(name)=="surfaceScalarField") {
        return true;
    } else {
        return false;
    }
}

bool ValueExpressionDriver::isFaceVector(const Foam::string &name)
{
    if(getTypeOfField(name)=="surfaceVectorField") {
        return true;
    } else {
        return false;
    }
}

bool ValueExpressionDriver::isScalar(const Foam::string &name)
{
    if(getTypeOfField(name)=="volScalarField") {
        return true;
    } else {
        return false;
    }
}

bool ValueExpressionDriver::isVector(const Foam::string &name)
{
    if(getTypeOfField(name)=="volVectorField") {
        return true;
    } else {
        return false;
    }
}

bool ValueExpressionDriver::isCellSet(const Foam::string &name)
{
    if(getTypeOfSet(name)=="cellSet") {
        return true;
    } else {
        return false;
    }
}

Foam::string ValueExpressionDriver::getTypeOfField(const Foam::string &name)
{
    Foam::IOobject f 
        (
            name,
            time,
            mesh,
            Foam::IOobject::MUST_READ,
            Foam::IOobject::NO_WRITE
        );
    f.headerOk();

    return f.headerClassName();
}

Foam::string ValueExpressionDriver::getTypeOfSet(const Foam::string &name)
{
    Foam::IOobject f 
        (
            name,
            "constant",
            Foam::polyMesh::meshSubDir/"sets",
	    mesh,
            Foam::IOobject::MUST_READ,
            Foam::IOobject::NO_WRITE
        );
    f.headerOk();

    return f.headerClassName();
}

Foam::volScalarField *ValueExpressionDriver::getScalarField(const Foam::string &name)
{
    Foam::dimensionSet nullDim(0,0,0,0,0);

    Foam::volScalarField *f=new Foam::volScalarField(
        Foam::IOobject
        (
            name,
            time,
            mesh,
            Foam::IOobject::MUST_READ,
            Foam::IOobject::NO_WRITE
        ),
        mesh
    );
    f->dimensions().reset(nullDim);

    return f;
}

Foam::surfaceScalarField *ValueExpressionDriver::getFaceScalarField(const Foam::string &name)
{
    Foam::dimensionSet nullDim(0,0,0,0,0);

    Foam::surfaceScalarField *f=new Foam::surfaceScalarField(
        Foam::IOobject
        (
            name,
            time,
            mesh,
            Foam::IOobject::MUST_READ,
            Foam::IOobject::NO_WRITE
        ),
        mesh
    );
    f->dimensions().reset(nullDim);

    return f;
}

Foam::surfaceVectorField *ValueExpressionDriver::getFaceVectorField(const Foam::string &name)
{
    Foam::dimensionSet nullDim(0,0,0,0,0);

    Foam::surfaceVectorField *f=new Foam::surfaceVectorField(
        Foam::IOobject
        (
            name,
            time,
            mesh,
            Foam::IOobject::MUST_READ,
            Foam::IOobject::NO_WRITE
        ),
        mesh
    );
    f->dimensions().reset(nullDim);

    return f;
}

Foam::volScalarField *ValueExpressionDriver::makeRandomField()
{
    Foam::volScalarField *f=makeScalarField(0.);
    Foam::Random rand(65);

    forAll(*f,cellI) {
        (*f)[cellI]=rand.scalar01();
    }

    return f;
}

Foam::volScalarField *ValueExpressionDriver::makeGaussRandomField()
{
    Foam::volScalarField *f=makeScalarField(0.);
    Foam::Random rand(65);

    forAll(*f,cellI) {
        (*f)[cellI]=rand.GaussNormal();
    }

    return f;
}

Foam::volVectorField *ValueExpressionDriver::makePositionField()
{
    Foam::dimensionSet nullDim(0,0,0,0,0);
    Foam::volVectorField *f=new Foam::volVectorField(
        Foam::IOobject
        (
            "pos",
            time,
            mesh,
            Foam::IOobject::NO_READ,
            Foam::IOobject::NO_WRITE
        ),
        mesh,
        Foam::vector(0,0,0)
    );
    f->dimensions().reset(mesh.C().dimensions());
    *f=mesh.C();
    f->dimensions().reset(nullDim);
    return f;
}

Foam::surfaceVectorField *ValueExpressionDriver::makeFacePositionField()
{
    Foam::dimensionSet nullDim(0,0,0,0,0);
    Foam::surfaceVectorField *f=new Foam::surfaceVectorField(
        Foam::IOobject
        (
            "fpos",
            time,
            mesh,
            Foam::IOobject::NO_READ,
            Foam::IOobject::NO_WRITE
        ),
        mesh,
        Foam::vector(0,0,0)
    );
    f->dimensions().reset(mesh.Cf().dimensions());
    *f=mesh.Cf();
    f->dimensions().reset(nullDim);
    return f;
}

Foam::surfaceVectorField *ValueExpressionDriver::makeFaceField()
{
    Foam::dimensionSet nullDim(0,0,0,0,0);
    Foam::surfaceVectorField *f=new Foam::surfaceVectorField(
        Foam::IOobject
        (
            "face",
            time,
            mesh,
            Foam::IOobject::NO_READ,
            Foam::IOobject::NO_WRITE
        ),
        mesh,
        Foam::vector(0,0,0)
    );
    f->dimensions().reset(mesh.Sf().dimensions());
    *f=mesh.Sf();
    f->dimensions().reset(nullDim);
    return f;
}

Foam::surfaceScalarField *ValueExpressionDriver::makeAreaField()
{
    Foam::dimensionSet nullDim(0,0,0,0,0);
    Foam::surfaceScalarField *f=new Foam::surfaceScalarField(
        Foam::IOobject
        (
            "face",
            time,
            mesh,
            Foam::IOobject::NO_READ,
            Foam::IOobject::NO_WRITE
        ),
        mesh,
        0
    );
    f->dimensions().reset(mesh.magSf().dimensions());
    *f=mesh.magSf();
    f->dimensions().reset(nullDim);
    return f;
}

Foam::volScalarField *ValueExpressionDriver::makeVolumeField()
{
    Foam::volScalarField *f=new Foam::volScalarField(
        Foam::IOobject
        (
            "vol",
            time,
            mesh,
            Foam::IOobject::NO_READ,
            Foam::IOobject::NO_WRITE
        ),
        mesh,
        0
    );
    const Foam::scalarField &V=mesh.V();

    forAll(*f,cellI) {
        (*f)[cellI]=V[cellI];
    }

    return f;
}

Foam::volScalarField *ValueExpressionDriver::makeDistanceField()
{
    Foam::dimensionSet nullDim(0,0,0,0,0);
    Foam::volScalarField *f=new Foam::volScalarField(
        Foam::IOobject
        (
            "dist",
            time,
            mesh,
            Foam::IOobject::NO_READ,
            Foam::IOobject::NO_WRITE
        ),
        mesh,
        0
    );
    f->dimensions().reset(mesh.C().dimensions());
    Foam::wallDist dist(mesh);
    *f=dist;
    f->dimensions().reset(nullDim);
    return f;

}

Foam::volVectorField *ValueExpressionDriver::getVectorField(const Foam::string &name)
{
    Foam::dimensionSet nullDim(0,0,0,0,0);
    Foam::volVectorField *f=new Foam::volVectorField(
        Foam::IOobject
        (
            name,
            time,
            mesh,
            Foam::IOobject::MUST_READ,
            Foam::IOobject::NO_WRITE
        ),
        mesh
    );

    f->dimensions().reset(nullDim);
    return f;
}

template<class T>
void ValueExpressionDriver::makeZeroGradientPatches(Foam::GeometricField<T,Foam::fvPatchField,Foam::volMesh> &field) 
{
  typename Foam::GeometricField<T,Foam::fvPatchField,Foam::volMesh>::GeometricBoundaryField &bf=field.boundaryField();
  Foam::List<Foam::fvPatchField<T> *>bfNew(bf.size());

  forAll(bf,patchI) {
    const Foam::fvPatch &patch=bf[patchI].patch();

    if(typeid(patch)==typeid(Foam::wallFvPatch) || typeid(patch)==typeid(Foam::fvPatch)) {
        bfNew[patchI]=new Foam::zeroGradientFvPatchField<T>(patch,field.internalField());      
    } else {
        bfNew[patchI]=bf[patchI].clone().ptr();
    }
  }

  bf.clear();
  bf.setSize(bfNew.size());
  forAll(bf,i) {
      bf.set(i)=bfNew[i];
  }
}

Foam::volScalarField *ValueExpressionDriver::makeScalarField(const Foam::scalar &val)
{
    std::ostringstream buff;
    buff << "constantScalar" << val;

    Foam::volScalarField *f=new Foam::volScalarField(
        Foam::IOobject
        (
            buff.str(),
            time,
            mesh,
            Foam::IOobject::NO_READ,
            Foam::IOobject::NO_WRITE
        ),
        mesh,
        val
    );
    makeZeroGradientPatches(*f);

    return f;
}

Foam::volScalarField *ValueExpressionDriver::makeCellSetField(const Foam::string &name)
{
  Foam::volScalarField *f=makeScalarField(0);

  Foam::cellSet cs(mesh,name);
  Foam::labelList cells(cs.toc());

  forAll(cells,cellI) {
    (*f)[cells[cellI]]=1.;
  }

  return f;
}

Foam::volVectorField *ValueExpressionDriver::makeVectorField(const Foam::vector &vec)
{
    std::ostringstream buff;
    buff << "constantVector" << vec.x() << "_" << vec.y() << "_" << vec.z() ;

    Foam::volVectorField *f=new Foam::volVectorField(
        Foam::IOobject
        (
            buff.str(),
            time,
            mesh,
            Foam::IOobject::NO_READ,
            Foam::IOobject::NO_WRITE
        ),
        mesh,
        vec
    );

    makeZeroGradientPatches(*f);

    return f;
}

Foam::volVectorField *ValueExpressionDriver::makeVectorField(Foam::volScalarField *x,Foam::volScalarField *y,Foam::volScalarField *z)
{
    Foam::volVectorField *f=makeVectorField(Foam::vector(0,0,0));

    forAll(*f,cellI) {
        (*f)[cellI]=Foam::vector((*x)[cellI],(*y)[cellI],(*z)[cellI]);
    }

    return f;
}

Foam::volScalarField *ValueExpressionDriver::doLogicalNot(Foam::volScalarField *a)
{
    Foam::volScalarField *res=makeScalarField(0.);

    forAll(*res,cellI) {
        bool av= (*a)[cellI] ? true : false;

        if( !av ) {
            (*res)[cellI]=1;
        } else {
            (*res)[cellI]=0;
        }
    }

    return res;
}

template<class T>
void ValueExpressionDriver::makePatches(Foam::GeometricField<T,Foam::fvPatchField,Foam::volMesh> &field,bool keepPatches,const Foam::wordList &fixedPatches) 
{
  typename Foam::GeometricField<T,Foam::fvPatchField,Foam::volMesh>::GeometricBoundaryField &bf=field.boundaryField();
  Foam::List<Foam::fvPatchField<T> *>bfNew(bf.size());

  forAll(bf,patchI) {
    const Foam::fvPatch &patch=bf[patchI].patch();

    if(!keepPatches && (typeid(patch)==typeid(Foam::wallFvPatch) || typeid(patch)==typeid(Foam::fvPatch))) {
        bool isValuePatch=false;
        forAll(fixedPatches,i) {
            if(fixedPatches[i]==patch.name()) {
                isValuePatch=true;
            }
        }
        if(isValuePatch){
            bfNew[patchI]=new Foam::fixedValueFvPatchField<T>(patch,field.internalField());  
        } else {
            bfNew[patchI]=new Foam::zeroGradientFvPatchField<T>(patch,field.internalField());      
        }
    } else {
        bfNew[patchI]=bf[patchI].clone().ptr();
    }
  }

  bf.clear();
  bf.setSize(bfNew.size());
  forAll(bf,i) {
      bf.set(i)=bfNew[i];
  }
}

template<class T>
void ValueExpressionDriver::setValuePatches(Foam::GeometricField<T,Foam::fvPatchField,Foam::volMesh> &field,bool keepPatches) 
{
  typename Foam::GeometricField<T,Foam::fvPatchField,Foam::volMesh>::GeometricBoundaryField &bf=field.boundaryField();
  Foam::List<Foam::fvPatchField<T> *>bfNew(bf.size());

  forAll(bf,patchI) {
    const Foam::fvPatch &patch=bf[patchI].patch();

    if(!keepPatches && (typeid(patch)==typeid(Foam::wallFvPatch) || typeid(patch)==typeid(Foam::fvPatch))) {
        if(typeid(field.boundaryField()[patchI])==typeid(Foam::fixedValueFvPatchField<T>)) {
            Foam::fvPatchField<T> &pf=field.boundaryField()[patchI];
            
            pf==pf.patchInternalField();
        }
    }
  }
}

// Force the compiler to generate the code, there'S a better way but I'm too stupid
void dummyS(Foam::GeometricField<Foam::scalar,Foam::fvPatchField,Foam::volMesh>  &f,bool keepPatches,const Foam::wordList &fixedPatches) {
    ValueExpressionDriver::makePatches(f,keepPatches,fixedPatches);
    ValueExpressionDriver::setValuePatches(f,keepPatches);
}

void dummyV(Foam::GeometricField<Foam::vector,Foam::fvPatchField,Foam::volMesh>  &f,bool keepPatches,const Foam::wordList &fixedPatches) {
    ValueExpressionDriver::makePatches(f,keepPatches,fixedPatches);
    ValueExpressionDriver::setValuePatches(f,keepPatches);
}
